import math

def default() -> dict:
    return {
        "strength": {
            "score": 10,
        },
        "dexterity": {
            "score": 10,
        },
        "constitution": {
            "score": 10,
        },
        "intelligence": {
            "score": 10,
        },
        "wisdom": {
            "score": 10,
        },
        "charisma": {
            "score": 10,
        },
    }

def calculate(a: dict, effect_total: int = 0):
    """Calculate a new ability score and modifier, and add them to the input dict."""
    base_value = a.get("score", 0)
    drain = a.get("drain", 0)

    modified_score = base_value - drain + effect_total
    a["modified_score"] = modified_score
    a["modifier"] = modifier(modified_score)

def modifier(score: int) -> int:
    """Calculates the appropriate ability modifier given an integer input representing an ability score."""
    calculated = (score * 0.5) - 5

    return int(math.floor(calculated))

def penalty(a: dict) -> int:
    damage = a.get("damage", 0)
    if damage == 0:
        return 0

    calculation = 0.5 * damage
    value = -int(math.floor(calculation))

    return value



from rerole_lib import ability
from rerole_lib import effect
from rerole_lib import save
from rerole_lib import skill
from rerole_lib import utils

def default() -> dict:
    """Create a new blank character sheet"""
    return {
        "abilities": ability.default(),
        "saves": save.default(),
        "skills": skill.default(),
    }

def calculate(data: dict) -> dict:
    """Calulate all relevant modifiers, returning a new dict."""
    antimagic_field_is_on = data.get("antimagic_field", False)
    if antimagic_field_is_on:
        activate_antimagic_field(data)
    else:
        deactivate_antimagic_field(data)

    update_effect_index(data)
    effect_index = data.get("effect_index", {})

    for k, v in data.get("abilities", {}).items():
        ability_effects = resolve_effect_index(data, k)
        effect_total = effect.total(ability_effects)
        ability.calculate(v, effect_total)

    for k, v in data.get("saves", {}).items():
        save_effects = resolve_effect_index(data, k)
        save_effect_total = effect.total(save_effects)

        save_ability = utils.get_in(data, ["abilities", v.get("ability")], {})
        save_ability_modifier = save_ability.get("modifier", 0)
        save_ability_penalty = ability.penalty(save_ability)

        effect_total = save_effect_total + save_ability_modifier + save_ability_penalty
        save.calculate(v, effect_total)

    for k, v in data.get("skills", {}).items():
        skill_effects = resolve_effect_index(data, k)
        skill_effect_total = effect.total(skill_effects)

        skill_ability = utils.get_in(data, ["abilities", v.get("ability")], {})
        skill_ability_modifier = skill_ability.get("modifier", 0)
        skill_ability_penalty = ability.penalty(skill_ability)

        effect_total = skill_effect_total + skill_ability_modifier + skill_ability_penalty
        skill.calculate(v, effect_total)


def activate_antimagic_field(data: dict):
    """Apply the proper suppression state to each magical effect present."""
    active_magic_effect_key_seqs = utils.search(data, active_magic_effect)
    if not active_magic_effect_key_seqs:
        return

    for seq in active_magic_effect_key_seqs:
        e = utils.get_in(data, seq, {})
        effect.toggle_antimagic_field(e)

def deactivate_antimagic_field(data: dict):
    """Like activate_antimagic_field, but in reverse."""
    inactive_magic_effect_key_seqs = utils.search(data, inactive_magic_effect)
    if not inactive_magic_effect_key_seqs:
        return

    for seq in inactive_magic_effect_key_seqs:
        e = utils.get_in(data, seq, {})
        effect.toggle_antimagic_field(e)


def update_effect_index(data: dict):
    """Add an up-to-date effect index to the provided character dict."""
    effect_index = build_effect_index(data)
    if not effect_index:
        return

    data["effect_index"] = effect_index

def build_effect_index(data: dict) -> dict:
    """Finds all effects in character data, and builds an index of things->effect key sequences.

    This function assumes that names of things are globally unique. If a character has an ability called 'strength' and a skill called 'strength', the resulting effect index will squish them together into a single entry.

    In practice, things which have effects applied to them generally have globally unique names, as they're things like abilities, saving throws, skills, and various built-in rolls, like AC and spellcasting concentration checks.
    """
    effects = utils.search(data, lambda x: isinstance(x, dict) and "affects" in x.keys())

    if not effects:
        return {}

    effect_index = {}
    for key_seq in effects:
        effect = utils.get_in(data, key_seq)
        if not effect:
            continue

        affecting_rules = effect["affects"]

        for group, value in affecting_rules.items():
            type_ = type(value)
            if type_ is bool and value is True:
                for item in data.get(group, {}).keys():
                    utils.add_or_append(effect_index, item, key_seq)
            elif type_ is str:
                utils.add_or_append(effect_index, value, key_seq)
            elif type_ is list:
                for item in value:
                    utils.add_or_append(effect_index, item, key_seq)
            else:
                continue

    return effect_index

def resolve_effect_index(data: dict, name: str) -> list[dict]:
    """Return a list of effects that are affecting the named item."""
    effect_key_seqs = utils.get_in(data, ["effect_index", name])
    if not effect_key_seqs:
        return []

    return [utils.get_in(data, seq) for seq in effect_key_seqs]


def active_magic_effect(e: dict) -> bool:
    return isinstance(e, dict) and effect.active(e) and e.get("magic", False)

def inactive_magic_effect(e: dict) -> bool:
    return isinstance(e, dict) and effect.inactive(e) and e.get("magic", False)



def default() -> dict:
    return {
        "acrobatics": {
            "ability": "dexterity",
        },
        "appraise": {
            "ability": "intelligence",
        },
        "bluff": {
            "ability": "charisma",
        },
        "climb": {
            "ability": "strength",
        },
        "craft": {
            "ability": "intelligence",
        },
        "diplomacy": {
            "ability": "charisma",
        },
        "disable device": {
            "ability": "dexterity",
        },
        "disguise": {
            "ability": "charisma",
        },
        "escape artist": {
            "ability": "dexterity",
        },
        "fly": {
            "ability": "dexterity",
        },
        "handle animal": {
            "ability": "charisma",
        },
        "heal": {
            "ability": "wisdom",
        },
        "intimidate": {
            "ability": "charisma",
        },
        "knowledge (arcana)": {
            "ability": "intelligence",
        },
        "knowledge (dungeoneering)": {
            "ability": "intelligence",
        },
        "knowledge (engineering)": {
            "ability": "intelligence",
        },
        "knowledge (geography)": {
            "ability": "intelligence",
        },
        "knowledge (history)": {
            "ability": "intelligence",
        },
        "knowledge (local)": {
            "ability": "intelligence",
        },
        "knowledge (nature)": {
            "ability": "intelligence",
        },
        "knowledge (nobility)": {
            "ability": "intelligence",
        },
        "knowledge (planes)": {
            "ability": "intelligence",
        },
        "knowledge (religion)": {
            "ability": "intelligence",
        },
        "linguistics": {
            "ability": "intelligence",
        },
        "perception": {
            "ability": "wisdom",
        },
        "perform": {
            "ability": "charisma",
        },
        "profession": {
            "ability": "wisdom",
        },
        "ride": {
            "ability": "dexterity",
        },
        "sense motive": {
            "ability": "wisdom",
        },
        "sleight of hand": {
            "ability": "dexterity",
        },
        "spellcraft": {
            "ability": "intelligence",
        },
        "stealth": {
            "ability": "dexterity",
        },
        "survival": {
            "ability": "wisdom",
        },
        "swim": {
            "ability": "strength",
        },
        "use magic device": {
            "ability": "charisma",
        },
    }

def calculate(s: dict, effect_total: int = 0):
    ranks = s.get("ranks", 0)
    is_class = s.get("class", False)
    has_ranks = ranks > 0

    class_bonus = 3 if has_ranks and is_class else 0

    modifier = ranks + class_bonus + effect_total

    s["modifier"] = modifier



from enum import Enum

def total(effects: list[dict]) -> int:
    """Calculate the total value of the provided effects."""
    applied_effects = applied(effects)
    if not applied_effects:
        return 0

    values = [b['value'] for b in applied_effects]

    return sum(values)

def applied(effects: list[dict]) -> list[dict]:
    """Filter out effects that do not apply due to stacking rules.

    Bonus stacking rules are not that complicated:

    * Bonuses/penalties of the same type generally do not stack, with some exceptions
    * The largest absolute value of each type of bonus and penalty will be applied
    """
    if not effects:
        return []

    active_effects = list(filter(active, effects))
    penalties = list(filter(penalty, active_effects))
    bonuses = list(filter(bonus, active_effects))

    non_stacking_penalties = list(filter(non_stacking, penalties))
    non_stacking_bonuses = list(filter(non_stacking, bonuses))

    non_stacking_penalties_by_type = {}
    for p in non_stacking_penalties:
        t = p.get('type', type.UNTYPED)
        if t not in non_stacking_penalties_by_type.keys():
            non_stacking_penalties_by_type[t] = []
        non_stacking_penalties_by_type[t].append(p)

    largest_non_stacking_penalties = []
    # Sorting penalties by 'value' gives us the most negative value first
    for _, Ps in non_stacking_penalties_by_type.items():
        Ps_by_size = sorted(Ps, key=lambda p: p['value'])
        largest_P = Ps_by_size[0]
        largest_non_stacking_penalties.append(largest_P)

    non_stacking_bonuses_by_type = {}
    for b in non_stacking_bonuses:
        t = b.get('type', type.UNTYPED)
        if t not in non_stacking_bonuses_by_type.keys():
            non_stacking_bonuses_by_type[t] = []
        non_stacking_bonuses_by_type[t].append(b)

    largest_non_stacking_bonuses = []
    # Sorting bonuses by 'value' _descending_ gives us the largest value first
    for _, Bs in non_stacking_bonuses_by_type.items():
        Bs_by_size = sorted(Bs, key=lambda b: b['value'], reverse=True)
        largest_B = Bs_by_size[0]
        largest_non_stacking_bonuses.append(largest_B)

    stacking_penalties = list(filter(stacking, penalties))
    stacking_bonuses = list(filter(stacking, bonuses))

    return stacking_penalties + stacking_bonuses + largest_non_stacking_penalties + largest_non_stacking_bonuses

class type(str, Enum):
    UNTYPED = "untyped"
    ALCHEMICAL = "alchemical"
    ARMOR = "armor"
    BAB = "bab"
    CIRCUMSTANCE = "circumstance"
    COMPETENCE = "competence"
    DEFLECTION = "deflection"
    DODGE = "dodge"
    ENHANCEMENT = "enhancement"
    INHERENT = "inherent"
    INSIGHT = "insight"
    LUCK = "luck"
    MORALE = "morale"
    NATURAL_ARMOR = "natural armor"
    PROFANE = "profane"
    RACIAL = "racial"
    RESISTANCE = "resistance"
    SACRED = "sacred"
    SHIELD = "shield"
    SIZE = "size"
    TRAIT = "trait"


stacking_types = [type.CIRCUMSTANCE, type.DODGE, type.UNTYPED, None]
active_states = ["active", None]
inactive_states = ["inactive", "suppressed", "disabled"]
permanent_states = ["suppressed", None]

"""
"active" and "inactive" indicate the status of a togglable effect.

"suppressed" indicates a temporarily suppressed permanent effect.

"disabled" indicates a temporarily suppressed togglable effect.

An effect without a "state" field is assumed to be permanently active.
"""

antimagic_field_state = {
    "suppressed": None,
    None: "suppressed",
    "active": "disabled",
    "disabled": "active",
    "inactive": "inactive",
}

def toggle_antimagic_field(e: dict):
    change_to = antimagic_field_state.get(e.get("state"))
    if change_to is None and "state" in e.keys():
        del e["state"]
        return
    e["state"] = change_to

def permanent(e: dict) -> bool:
    return e.get("state") in permanent_states

def togglable(e: dict) -> bool:
    return not permanent(e)

def active(e: dict) -> bool:
    return e.get("state") in active_states

def inactive(e: dict) -> bool:
    return not active(e)

def stacking(e: dict) -> bool:
    t = e.get("type")
    return t in stacking_types

def non_stacking(e: dict) -> bool:
    return not stacking(e)

def penalty(e: dict) -> bool:
    return e["value"] < 0

def bonus(e: dict) -> bool:
    return not penalty(e)



def default() -> dict:
    return {
        "fortitude": {
            "ability": "constitution",
        },
        "reflex": {
            "ability": "dexterity",
        },
        "will": {
            "ability": "wisdom",
        },
    }

def calculate(s: dict, effect_total: int):
    modifier = s.get("value", 0) + effect_total
    s["modifier"] = modifier
